
/*
 * This file is a sub module, H_elim.v, which generates the reduced echelon form of parity check matrix H.
 *
 * Copyright (C) 2017
 * Authors: Wen Wang <wen.wang.ww349@yale.edu>
 *          Ruben Niederhagen <ruben@polycephaly.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 *
*/


module H_elim
#(
  parameter N = 40,
  parameter L = 1547,
  parameter K = 8192,
  parameter M_elim = 1,
  parameter m = 13,
  parameter t = 119,
  parameter BLOCK = 20,
  // local:
  parameter K_elim = N*((K+N-1)/N),
  parameter mem_width = 1 << (m-`CLOG2(t+1)-1) // actual memory width in add_FFT, width = 2*mem_width*m
)
(
  input wire clk,
  input wire rst,
  input wire [m*(t+1)-1:0] coeff_in,
  input wire FYS_start,
  input wire eva_start,
  input wire [63:0] seed,
  input wire rd_en,
  input wire [`CLOG2(L*K_elim/N) - 1 : 0] rd_addr,
  input wire host_P_rd_en,
  input wire [m-1:0] host_P_rd_addr,
  input wire key_gen_done,
  
  output wire elim_done,
  output wire [N*M_elim-1:0] elim_data_out,
  output wire [m-1:0] P_dout,
  output wire [2*m*mem_width-1:0] eva_dout,
  output wire fail,
  
  // added for testing purpose
  output wire init_done,
  output wire FYS_done_out,
  output wire eva_done_out
);

wire mem_wr_en;
wire [`CLOG2(L*K_elim/N) - 1 : 0] mem_wr_addr;
wire [N*M_elim-1:0] mem_data_in;

reg wr_en_buffer = 0;
reg [`CLOG2(L*K_elim/N) - 1 : 0] wr_addr_buffer = 0;
reg [N*M_elim-1:0] data_in_buffer = 0;
reg GS_start = 0;

reg init_done_running_buffer = 0;

gen_H #(.N(N), .L(L), .K(K), .M_elim(M_elim), .m(m), .t(t)) gen_H_inst (
  .clk(clk),
  .rst(rst),
  .FYS_start(FYS_start),
  .eva_start(eva_start),
  .coeff_in(coeff_in),
  .seed(seed),
  .mem_wr_en(mem_wr_en),
  .mem_wr_addr(mem_wr_addr),
  .mem_data_in(mem_data_in),
  .init_done(init_done),
  .FYS_done_out(FYS_done_out),
  .eva_done_out(eva_done_out),
  .P_dout(P_dout),
  .eva_dout(eva_dout),
  .host_P_rd_en(host_P_rd_en),
  .host_P_rd_addr(host_P_rd_addr),
  .key_gen_done(key_gen_done)
);
 
wire init_done_running = init_done | init_done_running_buffer;
 always @(posedge clk)
  begin
    init_done_running_buffer <= FYS_start ? 0 :
                                init_done ? 1'b1 : init_done_running_buffer;
  end

 
always @(posedge clk)
  begin
    GS_start <= init_done;
    wr_en_buffer <= ((!init_done_running) && mem_wr_en);
    wr_addr_buffer <= mem_wr_addr;
    data_in_buffer <= mem_data_in;
  end
 
elim_H_elim #(.N(N), .L(L), .K(K_elim), .M(M_elim), .BLOCK(BLOCK), .DATA("")) elim_inst (
  .clk(clk),
  .rst(rst),
  .start(GS_start),
  .done(elim_done),
  .fail(fail),
  .rd_en(rd_en),
  .rd_addr(rd_addr),
  .data_out(elim_data_out),
  .wr_en(wr_en_buffer),
  .wr_addr(wr_addr_buffer),
  .data_in(data_in_buffer)
);

endmodule

