
/*
 * This file is a sub module, gen_H.v, which generates the parity check matrix H.
 *
 * Copyright (C) 2017
 * Authors: Wen Wang <wen.wang.ww349@yale.edu>
 *          Ruben Niederhagen <ruben@polycephaly.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 *
*/


module gen_H
#(
  parameter N = 20,
  parameter K = 6960,
  parameter M_elim = 1,
  parameter m = 13,
  parameter t = 119,
  parameter L = 1547,
  // local:
  parameter K_elim = N*((K+N-1)/N),
  parameter mem_width = 1 << (m-`CLOG2(t+1)-1)
)
(
  input wire clk,
  input wire rst,
  input wire FYS_start,
  input wire eva_start,
  input wire [m*(t+1)-1:0] coeff_in,
  input wire [63:0] seed,
  input wire host_P_rd_en,
  input wire [m-1:0] host_P_rd_addr,
  input wire key_gen_done,
  
  output wire mem_wr_en,
  output wire [`CLOG2(L*K_elim/N)-1 : 0] mem_wr_addr,
  output wire [N*M_elim-1:0] mem_data_in,
  output wire init_done,
// added for testing purpose
  output wire FYS_done_out,
  output wire eva_done_out,
  
  output wire [m-1:0] P_dout,
  output wire [2*m*mem_width-1:0] eva_dout
);
  
wire FYS_done;
wire eva_done;

 // added for testing purpose
assign FYS_done_out = FYS_done;
assign eva_done_out = eva_done;

 
wire inv_start;

wire [`CLOG2(K_elim/N)-1:0] block;
wire P_rd_en;
wire [m-1:0] P_rd_addr;
wire eva_rd_en;
wire [`CLOG2(t+1)-1:0] eva_rd_addr;
wire [N*m-1:0] index_in;
wire [N*m-1:0] eva_inv_in;
wire inv_done;

reg running  = 0;
always @(posedge clk)
  begin
    running <= eva_start ? 1'b1 :
              (block > (K_elim/N-1)) ? 1'b0 : running;
  end

wire block_init_done;

reg inv_start_buffer = 0;
always @(posedge clk)
  begin
    inv_start_buffer <= eva_start ? 0 :
                       (eva_done | (block_init_done && (block < (K_elim/N-1)))) ? 1'b1 : 0;  
  end
assign inv_start = inv_start_buffer;
	  
reg [`CLOG2(K_elim/N)-1:0] block_buffer = 0;
always @(posedge clk)
  begin
    block_buffer <= eva_start ? 0 :
                    (block_init_done && (block <= (K_elim/N-1))) ? block_buffer + 1 : block_buffer;
  end
assign block = block_buffer;
	  	  
reg init_done_buffer = 0;
always @(posedge clk)
  begin
    init_done_buffer <= eva_start ? 1'b0 : (block_init_done && (block == (K_elim/N-1)));
  end
assign init_done = init_done_buffer;

inv_g_alpha #(.N(N), .L(L), .K(K), .M_elim(M_elim), .m(m), .t(t)) inv_g_alpha_inst (
  .clk(clk),
  .start(inv_start),
  .P_dout(P_dout),
  .eva_dout(eva_dout),
  .block(block),
  .P_rd_en(P_rd_en),
  .P_rd_addr(P_rd_addr),
  .eva_rd_en(eva_rd_en),
  .eva_rd_addr(eva_rd_addr),
  .index_out(index_in),
  .inv_out(eva_inv_in),
  .done(inv_done)
);
 
gen_H_entries #(.N(N), .L(L), .K(K_elim), .M_elim(M_elim), .m(m)) gen_H_entries_inst (
	.clk(clk),
	.start(inv_done),
	.block(block),
	.index(index_in),
	.eva_inv(eva_inv_in),
	.wr_en(mem_wr_en),
	.data_in(mem_data_in),
	.wr_addr(mem_wr_addr),
	.block_init_done(block_init_done) 
);

wire FYS_INIT_done;
wire FYS_rd_en;
wire [m-1:0] FYS_rd_addr;
  
assign FYS_rd_en = !key_gen_done ? P_rd_en : host_P_rd_en;
assign FYS_rd_addr = !key_gen_done ? P_rd_addr : host_P_rd_addr;
  
FYS #(.m(m)) FYS_inst (
	.clk(clk),
	.start(FYS_start),
	.seed(seed),
  .rd_en(FYS_rd_en),
  .rd_addr_P(FYS_rd_addr),
	.INIT_done(FYS_INIT_done),
	.done(FYS_done),
	.data_out(P_dout)
);
 
  add_FFT #(.numc(t+1), .gf(m), .mem_width(mem_width)) add_FFT_inst (
	.clk(clk),
	.start(eva_start),
	.coeff_in(coeff_in),
	.rd_en(eva_rd_en),
	.rd_addr(eva_rd_addr),
 	.data_out(eva_dout),
	.done(eva_done) 
);
 
endmodule

