##
 # This file generates the theoretical key pair.
 #
 # Copyright (C) 2017
 # Authors: Wen Wang <wen.wang.ww349@yale.edu>
 #          Ruben Niederhagen <ruben@polycephaly.org>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
 # the Free Software Foundation; either version 3 of the License, or
 # (at your option) any later version.
 #
 # This program is distributed in the hope that it will be useful,
 # but WITHOUT ANY WARRANTY; without even the implied warranty of
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 # GNU General Public License for more details.
 #
 # You should have received a copy of the GNU General Public License
 # along with this program; if not, write to the Free Software Foundation,
 # Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 #
##

import sys

import argparse

parser = argparse.ArgumentParser(description='Generate memory contents for H.',
                formatter_class=argparse.ArgumentDefaultsHelpFormatter)
 
parser.add_argument('-t', dest='t', type=int, default=119,  
          help='number of errors; default = 119')
parser.add_argument('-N', dest='N', type=int, default=20,  
          help='column block size') 
parser.add_argument('-n', dest='n', type=int, default=8192,  
          help='code length') 
parser.add_argument('-m', '--gf', dest='m', type=int, default=13,
          help='field GF(2^m); default = GF(2^13)')
parser.add_argument('in_coeff', metavar='in_coeff', type=str,
                    help='input file containing coefficients')
parser.add_argument('in_P', metavar='in_P', type=str,
                    help='input file containing permuted indices')
 
args = parser.parse_args()

t = args.t
m = args.m
N = args.N
n = args.n
K = 2^m
L = m*t
n_ceil = N*int((n-1+N)/N)

F.<x> = GF(2)[]
KK.<a> = GF(2^m, modulus="first_lexicographic")
Kx.<x> = KK[]
 
fmt = "{0:0" + str(m) + "b}"

with open(args.in_coeff, "r") as f:
  coeff_in = f.readline().strip()

coeff = []

while coeff_in != "":
  coeff.append(Kx(KK.fetch_int(int(coeff_in[:m], 2))))
  coeff_in = coeff_in[m:]

# reconstruct the polynomial r(x)
r = Kx.zero()

for i, v in enumerate(reversed(coeff)):
  r += x^i*v

#print r

def random_irred(K, t, r):

  iden_matrix = matrix(K, t, t)
  for i in range(t):
      iden_matrix[i, i] = 1
  
  coeff = matrix(K, t, t + 1)
  
  # get irreducible polynomial
  g = PolynomialRing(GF(2), 'x').irreducible_element(t, algorithm="first_lexicographic")
  
  Kg = K.extension(g)
  ri = Kg.one()
    
  for i in range(t + 1):
      for j in range(t):
          coeff[j, i] = ri[j]
      ri *= r
    
  coeff_gs = coeff.echelon_form()
  coeff_gs_sq = coeff_gs[:, :t]

  r_min_poly = x^t
  for i in range(t):
      r_min_poly += coeff_gs[i, t]*x^i

  return r_min_poly

# generate random polynomial based on minimal polynomial method
g = random_irred(KK, t, r)

#for i in range(t):
#  print fmt.format(g[i].integer_representation())

# reconstruct permutation indices
P_list = []

with open(args.in_P, "r") as f:
  for line in f:
    if not line.startswith("//"):
      if not (line == "\n"):
        P_list.append(int(line[:m], 2))

#print len(P_list)
#print P_list

# polynomial evaluation
basis = [a^i for i in reversed(range(m))]

def convert(beta, num):
  ret = 0
  for i in range(len(beta)):
    if (num >> i) % 2:
      ret += beta[i]
  return ret

alpha = []

for j in range(2^m):
  alpha.append(KK(convert(basis, j)))

eva = []

for i in range(2^m):
  x = alpha[i]
  eva.append(g(x)) 
  if (g(x) == KK.zero()):
    print 'ERROR!', i, x
 
Y = matrix(KK, t, K)
for i in range(K):
  Y[0, i] = KK.one()
  for j in range(1, t):
    Y[j, i] = Y[j-1, i]*alpha[i]

#print Y[1, 1]

Z = matrix(KK, K, K)
for i in range(K):
  Z[i, i] = (eva[i])^-1
     
P_trans = matrix(KK, K, K)
 
for i in range(K):
  row = P_list[i]
  P_trans[row, i] = 1

H = Y*Z*P_trans

H_expand = matrix(KK, t, n_ceil)
for i in range(n_ceil):
  for j in range(t):
    if (i < n):
      H_expand[j, i] = H[j, i]
    else:
      H_expand[j, i] = KK.zero()

H_2 = matrix(GF(2), L, n_ceil)

for i in range(n_ceil):
  for j in range(t):
    element = (H_expand[j, i]).integer_representation()
    for k in range(m):
      row_num = m*j+k
      if (element >> k) & 1 == 1:
        H_2[row_num, i] = 1
      else:
        H_2[row_num, i] = 0
        
iden_matrix = matrix(KK, L, L)
for i in range(L):
    iden_matrix[i, i] = 1
    
H_2 = H_2.echelon_form()
if H_2[:, :L] != iden_matrix:
  print "Please try new seeds, H_2 can not be eliminated."

#for i in range(n_ceil/N):
for i in range(int(L/N), n_ceil/N):
  for j in range(m*t):
    for k in reversed(range(N)):
      column_num = N*i+k
      element = H_2[j, column_num]
      sys.stdout.write(str(element))
    sys.stdout.write("\n")



