/*
 * This file is the top-level module.
 *
 * Copyright (C) 2017
 * Authors: Wen Wang <wen.wang.ww349@yale.edu>
 *          Ruben Niederhagen <ruben@polycephaly.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 *
*/

module key_gen
#(
  parameter m = 13,                   // 2^m is the size of the field
  parameter t = 119,                  // t is the number of errors
  parameter N_R_elim = 2,             // N_R_elim is the size of the GF(2^13) systolic array
  parameter BLOCK_R_elim = 1,
  parameter N_H_elim = 20,            // N_H_elim is the size of the GF(2) systolic array
  parameter BLOCK_H_elim = 20,
  parameter K = 8192,                 // K is the code length
  parameter M_H_elim = 1,             // M_H_elim is the size of GF(2) field
  // local:
  parameter K_R_elim = N_R_elim*((t+1+N_R_elim-1)/N_R_elim),
  parameter K_H_elim = N_H_elim*((K+N_H_elim-1)/N_H_elim),  // K_H_elim: column number for H's memory
  parameter L = m*t // L: row number of H
)
(
  // mandatory interface
  input wire start, //start&rst
  input wire clk,
  input wire [63:0] seed_0,
  input wire [63:0] seed_1,
  
  output wire done,
  output wire fail,
  
  // host interface for testing
  input wire host_P_rd_en,
  input wire [m-1:0] host_P_rd_addr,
  output wire [m-1:0] P_dout,
  
  input wire host_R_elim_rd_en,
  input wire [`CLOG2(t*K_R_elim/N_R_elim)-1 : 0] host_R_elim_rd_addr,
  output wire [N_R_elim*m-1:0] R_elim_dout,
  
  input wire host_H_elim_rd_en,
  input wire [`CLOG2(L*K_H_elim/N_H_elim) - 1 : 0] host_H_elim_rd_addr,
  output wire [N_H_elim*M_H_elim-1:0] H_elim_dout,
  
  // added for tetsing purpose
  output wire r_gen_done,
  output wire R_elim_init_done,
  output wire R_elim_done,
  output wire H_elim_init_done,
  output wire FYS_done,
  output wire eva_done,
  output wire H_elim_done
);

  localparam mem_width = 1 << (m-`CLOG2(t+1)-1);
  
  wire [t*m-1:0] r_gen_coeff_out;
  
  r_gen #(.m(m), .t(t)) r_gen_inst(
    .start(start),
    .clk(clk),
    .seed(seed_0),
    .coeff_out(r_gen_coeff_out),
    .done(r_gen_done)
  );
   
  reg R_elim_start = 0;
  reg [m*t-1:0] R_elim_coeff_in = 0;
  reg R_elim_rd_en = 0;
  reg [`CLOG2(t*K_R_elim/N_R_elim)-1 : 0] R_elim_rd_addr = 0;
   
  // R_elim starts when a random element in (GF(2^m)^t) arrives
  
  always @(posedge clk)
    begin
      R_elim_start <= r_gen_done;
      R_elim_coeff_in <= r_gen_done ? r_gen_coeff_out : R_elim_coeff_in;
    end
  
  reg [`CLOG2(t):0] R_elim_rd_counter = 0;
  reg R_elim_rd_running = 0;
  
  // when R_elim ends, read out the last column as g0, g1, ..., g_(t-1)
  
  always @(posedge clk)
    begin
      R_elim_rd_running <= R_elim_start ? 0 :
                           R_elim_done ? 1 : R_elim_rd_running;
      R_elim_rd_counter <= R_elim_start ? 0 :
                          (R_elim_rd_running && (R_elim_rd_counter <= (t+1))) ? R_elim_rd_counter + 1 : R_elim_rd_counter;
      R_elim_rd_en <= (R_elim_rd_running && (R_elim_rd_counter < t));
      R_elim_rd_addr <= R_elim_done ? t*K_R_elim/N_R_elim-1 :
                        R_elim_rd_en ? R_elim_rd_addr - 1 : R_elim_rd_addr;
                   
    end
  
  reg gen_priv_poly_done = 0;
  reg [m*t-1:0] priv_poly = 0;
  reg R_elim_rd_en_reg = 0;
  always @(posedge clk)
    begin
      gen_priv_poly_done <= R_elim_done ? 0 : (R_elim_rd_counter == (t+1));
      R_elim_rd_en_reg <= R_elim_rd_en;
      priv_poly <= R_elim_done ? 0 :
                   R_elim_rd_en_reg ? ((priv_poly << m) ^ R_elim_dout[m*(N_R_elim-(K_R_elim-(t+1)))-1 : m*(N_R_elim-(K_R_elim-(t+1))-1)]) : priv_poly;
    end
  
  wire R_elim_rd_en_in;
  wire [`CLOG2(t*K_R_elim/N_R_elim)-1 : 0] R_elim_rd_addr_in;

  reg key_gen_done = 0;
  
  assign R_elim_rd_en_in = !key_gen_done ? R_elim_rd_en : host_R_elim_rd_en;
  assign R_elim_rd_addr_in = !key_gen_done ? R_elim_rd_addr : host_R_elim_rd_addr;
  
  R_elim #(.N(N_R_elim), .M(m), .L(t), .K(t+1), .BLOCK(BLOCK_R_elim)) R_elim_inst(
    .clk(clk),
    .start(R_elim_start),
    .rst(1'b0),
    .coeff_in(R_elim_coeff_in),
    .rd_en(R_elim_rd_en_in),
    .rd_addr(R_elim_rd_addr_in),
    .data_out(R_elim_dout),
    .done(R_elim_done),
    .init_done(R_elim_init_done)
  );
  
  reg [m*(t+1)-1:0] H_elim_coeff_in = 0;
  reg [m-1:0] eva_coeff_in_one = 1;
  
  reg FYS_start = 0;
  reg eva_start = 0;
  
  always @(posedge clk)
    begin
      H_elim_coeff_in <= gen_priv_poly_done ? {eva_coeff_in_one, priv_poly} : H_elim_coeff_in;
      FYS_start <= R_elim_start;
      eva_start <= gen_priv_poly_done;   
    end
  
  reg done_buffer = 1'b0;
  always @(posedge clk)
    begin
      done_buffer <= H_elim_done;
    end
  
  assign done = done_buffer;
  
  always @(posedge clk)
    begin
      key_gen_done <= start ? 0 :
                      done ? 1 : key_gen_done;
    end
 
  wire [2*m*mem_width-1:0] eva_dout;
  
  wire H_elim_rd_en;
  wire [`CLOG2(L*K_H_elim/N_H_elim) - 1 : 0] H_elim_rd_addr;
  
  assign H_elim_rd_en = !key_gen_done ? 0 : host_H_elim_rd_en;
  assign H_elim_rd_addr = !key_gen_done ? 0 : host_H_elim_rd_addr;
   
  H_elim #(.N(N_H_elim), .L(L), .K(K), .m(m), .M_elim(M_H_elim), .t(t), .BLOCK(BLOCK_H_elim)) H_elim_inst(
    .clk(clk),
    .rst(1'b0),
    .coeff_in(H_elim_coeff_in),
    .FYS_start(FYS_start),
    .eva_start(eva_start),
    .seed(seed_1),
    .rd_en(H_elim_rd_en),
    .rd_addr(H_elim_rd_addr),
    .key_gen_done(key_gen_done),
    .host_P_rd_en(host_P_rd_en),
    .host_P_rd_addr(host_P_rd_addr),
    .init_done(H_elim_init_done),
    .FYS_done_out(FYS_done),
    .eva_done_out(eva_done),
    .elim_done(H_elim_done),
    .elim_data_out(H_elim_dout),
    .P_dout(P_dout),
    .eva_dout(eva_dout),
    .fail(fail)
  );
  
endmodule

  
  
  
  
   
  
